import tensorflow as tf


def create_padding_mask(seq):
    seq = tf.cast(tf.math.equal(seq, 0), tf.float32)

    # アテンション・ロジットにパディングを追加するため
    # さらに次元を追加する
    return seq[:, tf.newaxis, tf.newaxis, :]    # (batch_size, 1, 1, seq_len)


def create_look_ahead_mask(size):
    return 1 - tf.linalg.band_part(tf.ones((size, size)), -1, 0)    # (seq_len, seq_len)



from .attention import scaled_dot_product_attention
import tensorflow as tf


class MultiHeadAttention(tf.keras.layers.Layer):
    def __init__(self, d_model, num_heads):
        super(MultiHeadAttention, self).__init__()

        self.num_heads = num_heads
        self.d_model = d_model

        assert d_model % self.num_heads == 0

        self.depth = d_model // self.num_heads

        self.wq = tf.keras.layers.Dense(d_model)
        self.wk = tf.keras.layers.Dense(d_model)
        self.wv = tf.keras.layers.Dense(d_model)

        self.dense = tf.keras.layers.Dense(d_model)

    def split_heads(self, x, batch_size):
        """
        最後の次元を(num_heads, depth)に分割。
        結果をshapeが(batch_size, num_heads, seq_len, depth)となるようにリシェイプする。
        """
        x = tf.reshape(x, (batch_size, -1, self.num_heads, self.depth))
        return tf.transpose(x, perm=[0, 2, 1, 3])

    def call(self, v, k, q, mask):
        batch_size = tf.shape(q)[0]

        q = self.wq(q)  # (batch_size, seq_len, d_model)
        k = self.wk(k)  # (batch_size, seq_len, d_model)
        v = self.wv(v)  # (batch_size, seq_len, d_model)

        q = self.split_heads(q, batch_size)     # (batch_size, num_heads, seq_len_q, depth)
        k = self.split_heads(k, batch_size)     # (batch_size, num_heads, seq_len_k, depth)
        v = self.split_heads(v, batch_size)     # (batch_size, num_heads, seq_len_v, depth)

        # scaled_attention.shape == (batch_size, num_heads, seq_len_q, depth)
        # attention_weights.shape == (batch_size, num_heads, seq_len_q, seq_len_k)
        scaled_attention, attention_weights = scaled_dot_product_attention(q, k, v, mask)

        # (batch_size, seq_len_q, num_heads, depth)
        scaled_attention = tf.transpose(scaled_attention, perm=[0, 2, 1, 3])
        # (batch_size, seq_len_q, d_model)
        concat_attention = tf.reshape(scaled_attention, (batch_size, -1, self.d_model))

        output = self.dense(concat_attention)   # (batch_size, seq_len_q, d_model)

        return output, attention_weights



from mask import *
from multi_head_attention import *
from transformer import *


__copyright = 'Copyright (C) 2020 Otsuka Kentaro'
__version__ = '0.0.4'
__license__ = 'MIT License'
__author__ = 'Otsuka Kentaro'
__author_email__ = 'otsuka.kenchan@gmail.com'
__url__ = 'https://github.com/otsuka-kentaro/tensorflow-modules'



from . import util
import tensorflow as tf
import unittest


class TestMask(unittest.TestCase):
    def test_create_padding_mask(self):
        seq = tf.random.uniform([3, 5])

        mask = util.create_padding_mask(seq)
        self.assertEqual([3, 1, 1, 5], mask.shape)

    def test_create_padding_mask2(self):
        seq = tf.random.uniform([3, 4, 5])

        mask = util.create_padding_mask(seq)
        self.assertEqual([3, 1, 1, 4, 5], mask.shape)

    def test_create_look_ahead_mask(self):
        mask = util.create_look_ahead_mask(5)
        self.assertEqual([5, 5], mask.shape)
